import logging, json, nre, sequtils, strtabs, tables
import private/util

# Types
#
type
  Step* = object
    name*, stepCmd*, workingDir*: string
    artifacts*, cmdInput*, depends*, expectedEnv*: seq[string]
    dontCache*: bool

  ProjectCfg* = object
    name*: string
    versionCmd*: string
    steps*: Table[string, Step]

  ProjectDef* = object
    cfgFilePath*, defaultBranch*, name*, repo*: string
    envVars*: StringTableRef

  StrawBossConfig* = object
    artifactsRepo*: string
    projects*: seq[ProjectDef]


# internal utils

let nullNode = newJNull()
proc getIfExists(n: JsonNode, key: string): JsonNode =
  result = if n.hasKey(key): n[key]
           else: nullNode

# Configuration parsing code

proc loadStrawBossConfig*(cfgFile: string): StrawBossConfig =
  let jsonCfg = parseFile(cfgFile)

  var projectDefs: seq[ProjectDef] = @[]

  for pJson in jsonCfg.getIfExists("projects").getElems:
    if not pJson.hasKey("name"):
      raiseEx "a project definition is missing the project name"

    if not pJson.hasKey("repo"):
      raiseEx "a project definition is missing the project repo configuration"

    var envVars = newStringTable(modeCaseSensitive)
    for k, v in pJson.getIfExists("envVars").getFields: envVars[k] = v.getStr("")

    projectDefs.add(
      ProjectDef(
        cfgFilePath: pJson.getIfExists("cfgFilePath").getStr("strawboss.json"),
        defaultBranch: pJson.getIfExists("defaultBranch").getStr("master"),
        name: pJson["name"].getStr,
        envVars: envVars,
        repo: pJson["repo"].getStr))

  result = StrawBossConfig(
    artifactsRepo: jsonCfg.getIfExists("artifactsRepo").getStr("artifacts"),
    projects: projectDefs)

proc loadProjectConfig*(cfgFile: string): ProjectCfg =
  let jsonCfg = parseFile(cfgFile)

  if not jsonCfg.hasKey("name"):
    raise newException(Exception, "project configuration is missing a name")

  if not jsonCfg.hasKey("steps"):
    raise newException(Exception, "project configuration is missing steps definition")

  var steps = initTable[string, Step]()
  for sName, pJson in jsonCfg["steps"].getFields:
    steps[sName] = Step(
      name:         sName,
      workingDir:   pJson.getIfExists("workingDir").getStr("."),
      stepCmd:      pJson.getIfExists("stepCmd").getStr("sh"),
      depends:      pJson.getIfExists("depends").getElems.mapIt(it.getStr),
      artifacts:    pJson.getIfExists("artifacts").getElems.mapIt(it.getStr),
      cmdInput:     pJson.getIfExists("cmdInput").getElems.mapIt(it.getStr),
      expectedEnv:  pJson.getIfExists("expectedEnv").getElems.mapIt(it.getStr),
      dontCache:    pJson.getIfExists("dontCache").getStr("false") != "false")

    if steps[sName].stepCmd == "sh" and steps[sName].cmdInput.len == 0:
      warn "Step " & sName & " uses 'sh' as its command but has no cmdInput."

  result = ProjectCfg(
    name: jsonCfg["name"].getStr,
    versionCmd: jsonCfg.getIfExists("versionCmd").getStr("git describe --tags --always"),
    steps: steps)

