import logging, nre, os, osproc, sequtils, streams, strtabs, strutils, tables,
  tempfile

import private/util
import configuration

type
  Workspace* = ref object
    artifactsRepo*: string
    env*: StringTableRef
    workingDir*: string
    project*: ProjectCfg


let ENV = loadEnv()

let logProcOutput: HandleProcMsgCB = proc (cmd: string, outMsg: TaintedString, errMsg: TaintedString) =
  if outMsg != nil: info cmd & "(stdout): " & outMsg
  if errMsg != nil: info cmd & "(stderr): " & errMsg

let envVarRe = re"\$\w+|\$\{[^}]+\}"
proc resolveEnvVars(line: string, env: StringTableRef): string =
  result = line
  for found in line.findAll(envVarRe):
    let key = if found[1] == '{': found[2..^2] else: found[1..^1]
    if env.hasKey(key): result = result.replace(found, env[key])

let SB_EXPECTED_VARS = ["VERSION"]

proc setupProjectForWork(projectDef: ProjectDef, buildRef, artifactsRepo: string): Workspace =

  info "Setting up to do work for '" & projectDef.name & "' at ref " & buildRef & "."
  var env = ENV
  env["GIT_DIR"] = ".git"

  # Create a temp directory that we'll work in
  let projDir = mkdtemp()
  debug "Workspace for '" & projectDef.name & ":  " & projDir
  assert projDir.isAbsolute

  # Clone the project into the $temp/repo directory
  let cloneResult = exec("git", projDir, ["clone", projectDef.repo, "repo"],
    env, {poUsePath}, logProcOutput)

  if cloneResult.exitCode != 0:
    removeDir(projDir)
    raiseEx "unable to clone repo for '" & projectDef.name & "'"

  # Checkout the requested ref
  let checkoutResult = exec("git", projDir & "/repo", ["checkout", buildRef],
    env, {poUsePath}, logProcOutput)

  if checkoutResult.exitCode != 0:
    removeDir(projDir)
    raiseEx "unable to checkout ref " & buildRef & " for '" & projectDef.name & "'"

  # Find the strawboss project configuration
  let projCfgFile = projDir & "/repo/" & projectDef.cfgFilePath
  if not existsFile(projCfgFile):
    removeDir(projDir)
    raiseEx "Cannot find strawboss project configuration in the project " &
      "repo (expected at '" & projectDef.cfgFilePath & "')."

  let projectCfg = loadProjectConfig(projCfgFile)
  result = Workspace(env: env, workingDir: projDir, project: projectCfg,
                     artifactsRepo: artifactsRepo)

  # Merge in the project-defined env vars
  for k, v in projectDef.envVars: result.env[k] = v

  # Get the build version
  let versionProc = startProcess(
    projectCfg.versionCmd,      # command
    projDir & "/repo",          # working dir
    [],                         # args
    result.env,                 # environment
    {poUsePath, poEvalCommand}) # options

  let versionResult = waitForWithOutput(versionProc, logProcOutput,
                                        projectCfg.versionCmd)

  if versionResult.exitCode != 0:
    removeDir(projDir)
    raiseEx "Version command (" & projectCfg.versionCmd & ") returned non-zero exit code."

  debug "Building version " & versionResult.output.strip
  result.env["VERSION"] = versionResult.output.strip

proc runStep*(step: Step, wksp: Workspace): void =
  let stepArtifactDir = wksp.artifactsRepo & "/" & wksp.project.name & "/" & step.name & "/" & wksp.env["VERSION"]

  if existsDir(stepArtifactDir) and not step.dontCache:
    info "Skipping step '" & step.name & "': already completed."
    return

  info "Running step '" & step.name & "' for " & wksp.project.name

  # Ensure all expected environment variables are present.
  for k in (step.expectedEnv & @SB_EXPECTED_VARS):
    if not wksp.env.hasKey(k):
      debug "workspace.env = " & $(wksp.env)
      raiseEx "step " & step.name & " failed: missing required env variable: " & k

  # Ensure that artifacts in steps we depend on are present
  for dep in step.depends:
    if not wksp.project.steps.hasKey(dep):
      raiseEx step.name & " depends on " & dep & " but there is no step named " & dep
    let depStep = wksp.project.steps[dep]

    # Run that step (may get skipped)
    let depDir = wksp.artifactsRepo & "/" & wksp.project.name & "/" & dep & "/" & wksp.env["VERSION"]
    runStep(depStep, wksp)

    # Add the artifacts directory for the dependent step to our env so that
    # further steps can reference it via $<stepname>_DIR
    echo "FP: " & depDir
    wksp.env[depStep.name & "_DIR"] = depDir

  # Run the step command, piping in cmdInput
  debug step.name & ": starting stepCmd: " & step.stepCmd
  let cmdProc = startProcess(step.stepCmd, wksp.workingDir & "/repo/" & step.workingDir,
    [], wksp.env, {poUsePath, poEvalCommand})

  let cmdInStream = inputStream(cmdProc)

  # Replace env variables in step cmdInput as we pipe it in
  for line in step.cmdInput: cmdInStream.writeLine(line.resolveEnvVars(wksp.env))
  cmdInStream.flush()
  cmdInStream.close()
  
  let cmdResult = waitForWithOutput(cmdProc, logProcOutput, step.stepCmd)

  if cmdResult.exitCode != 0:
    raiseEx "step " & step.name & " failed: step command returned non-zero exit code"

  # Gather the output artifacts (if we have any)
  if not existsDir(stepArtifactDir): createDir(stepArtifactDir)
  if step.artifacts.len > 0:
    for a in step.artifacts:
      let artifactPath = a.resolveEnvVars(wksp.env)
      let artifactName = artifactPath[(artifactPath.rfind("/")+1)..^1]
      try:
        copyFile(wksp.workingDir & "/repo/" & step.workingDir & "/" & artifactPath,
                 stepArtifactDir & "/" & artifactName)
      except:
        removeDir(stepArtifactDir)
        raiseEx "step " & step.name & " failed: unable to copy artifact " &
          artifactPath & ":\n" & getCurrentExceptionMsg()

# TODO: change return to return logs, results, some compound, useful object.
proc runStep*(cfg: StrawBossConfig, projectName, stepName, buildRef: string): void =

  let matching = cfg.projects.filterIt(it.name == projectName)
  if matching.len == 0: raiseEx "no such project: " & projectName
  elif matching.len > 1: raiseEx "more than one project named : " & projectName

  let projectDef = matching[0]

  let foundBuildRef = if buildRef != nil: buildRef else: projectDef.defaultBranch
  let wksp = setupProjectForWork(projectDef, foundBuildRef, cfg.artifactsRepo)

  # Find the step
  if not wksp.project.steps.hasKey(stepName):
    raiseEx "no step name '" & stepName & "' for " & projectName

  let step = wksp.project.steps[stepName]

  runStep(step, wksp)

